# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division, print_function

from django import template

from push import settings

register = template.Library()


@register.simple_tag
def push_settings(name):
    return vars(settings)[name]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division, print_function

import pytz
from PIL import Image

import logging

from django.db import IntegrityError, connection
from django.http import JsonResponse, Http404, HttpResponse
from django.views.decorators.cache import never_cache, cache_page
from django.views.decorators.http import require_POST
from django.views.decorators.csrf import csrf_exempt, csrf_protect
from django.shortcuts import render, redirect
from django.conf import settings
from django.urls import reverse, NoReverseMatch
from django.core.validators import URLValidator
from django.core.exceptions import ValidationError
from django.contrib.staticfiles import finders
from django.contrib.sites.shortcuts import get_current_site

from .settings import FCM_SENDER_ID, GCM_URL, APP_ICON_URLS, USE_CSRF, \
                      APP_BACKGROUND_COLOR, APP_THEME_COLOR
from .models import DigestSubscription, Task, TimezoneLayout


logger = logging.getLogger(__name__)


@never_cache
def on_off(request):
    return render(request, 'push/on_off.html')


def info_disable(request):
    return render(request, 'push/info_disable.html')


@cache_page(86400)
def manifest_json(request):
    # we need this to set up gcm_sender_id
    # https://developers.google.com/web/updates/2014/11/Support-for-installable-web-apps-with-webapp-manifest-in-chrome-38-for-Android
    try:
        start_url = "%s" % reverse('homepage')
    except NoReverseMatch:
        start_url = "/"
    
    site = get_current_site(request)
    manifest = {
        "name": site.name,
        "short_name": site.name,
        "icons": [],
        "start_url": "%s?from=manifest_json" % start_url,
        "display": "standalone",
    }
    if FCM_SENDER_ID:
        manifest['gcm_sender_id'] = FCM_SENDER_ID
    
    for icon in APP_ICON_URLS:
        # we cache this for 24 hours, so using Pillow is acceptable
        img_file = finders.find(icon)
        with Image.open(img_file) as img:
            manifest['icons'].append({
                "src": settings.STATIC_URL + icon,
                "sizes": "%dx%d" % (img.width, img.height),
                "type": "image/%s" % img.format.lower()
            })
    
    if APP_BACKGROUND_COLOR is not None:
        manifest['background_color'] = APP_BACKGROUND_COLOR
    if APP_THEME_COLOR is not None:
        manifest['theme_color'] = APP_THEME_COLOR
    
    return JsonResponse(manifest)


def _save(request):
    try:
        ua = request.META['HTTP_USER_AGENT'][:255]
    except (AttributeError, KeyError):
        ua = ''
    
    endpoint = request.POST.get('endpoint')
    logger.debug('Push save endpoint: %s' % endpoint)
    try:
        url_validate = URLValidator(schemes=['https',])
        url_validate(endpoint)
    except ValidationError:
        # chrome 44 and 45 sometimes give only unique part not full endpoint
        # url, workaround
        if endpoint and ua and 'Chrome' in ua:
            endpoint = "%s/%s" % (GCM_URL, endpoint)
            logger.debug('Fixed endpoint to: %s' % endpoint)
        else:
            # endpoint must be an url
            raise Http404('Wrong endpoint')
    
    key = request.POST.get('key', '')
    auth_secret = request.POST.get('auth_secret', '')
    
    timezone = request.POST.get('timezone')
    if timezone not in pytz.all_timezones:
        timezone = settings.TIME_ZONE
    
    try:
        subscr = DigestSubscription.objects.get(endpoint=endpoint)
    except DigestSubscription.DoesNotExist:
        subscr = DigestSubscription()
        subscr.endpoint = endpoint
    # actualize info anyway
    subscr.key = key
    subscr.auth_secret = auth_secret
    subscr.timezone = timezone
    subscr.ua = ua
    subscr.reactivate_if_needed()
    subscr.save()
    
    return JsonResponse({
        'response': {
            'status': 'ok',
            'id': subscr.pk,
        }
    })


@require_POST
@csrf_exempt
def save(request):
    # https://docs.djangoproject.com/en/1.8/ref/csrf/#view-needs-protection-for-one-path
    @csrf_protect
    def _protected_save(request):
        return _save(request)
    
    if USE_CSRF:
        return _protected_save(request)
    else:
        return _save(request)


def _deactivate(request):
    endpoint = request.POST.get('endpoint')
    
    try:
        subscr = DigestSubscription.objects.get(endpoint=endpoint)
    except DigestSubscription.DoesNotExist:
        raise Http404
    subscr.deactivate().save()
    
    return JsonResponse({
        'response': {
            'status': 'ok',
            'id': subscr.pk,
        }
    })


@require_POST
@csrf_exempt
def deactivate(request):
    # https://docs.djangoproject.com/en/1.8/ref/csrf/#view-needs-protection-for-one-path
    @csrf_protect
    def _protected_deactivate(request):
        return _deactivate(request)
    
    if USE_CSRF:
        return _protected_deactivate(request)
    else:
        return _deactivate(request)


def _notification_plus_one(what, id):
    """
    Statistics function (using raw SQL for speed).
    """
    cursor = connection.cursor()
    # what variable content is controlled in django routing
    # no security threats here
    cursor.execute("UPDATE %s SET %s=%s+1 WHERE id=%d" \
                    % (Task._meta.db_table, what, what, int(id)))


@never_cache
def notification_plus_one(request, what, id):
    """
    View to count views/closings statistics for push task.
    """
    _notification_plus_one(what, id)
    return HttpResponse(content='ok')


@never_cache
def last_notification(request):
    """
    Payload of the last actual and active push task for default timezone
    for old legacy push subscriptions what do not support payload encryption.
    """
    try:
        tz = TimezoneLayout.public_objects.filter(timezone=settings.TIME_ZONE)[0]
    except IndexError:
        raise Http404
    _notification_plus_one('views', tz.task.id)
    return JsonResponse({ 'notification': tz.task.get_payload(), })


@never_cache
def show_notification(request, id):
    try:
        task = Task.public_objects.get(pk=id)
    except Task.DoesNotExist:
        raise Http404
    _notification_plus_one('clicks', task.id)
    return redirect(task.url_relative(add_querystring={'from': 'push'}))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division, print_function

from django.conf import settings

VAPID_PUBLIC_KEY = getattr(settings, 'DJANGO_INFOPUSH_VAPID_PUBLIC_KEY', '')
VAPID_PRIVATE_KEY = getattr(settings, 'DJANGO_INFOPUSH_VAPID_PRIVATE_KEY', '')
VAPID_ADMIN_EMAIL = getattr(settings, 'DJANGO_INFOPUSH_VAPID_ADMIN_EMAIL', '')

FCM_SERVER_KEY = getattr(settings, 'DJANGO_INFOPUSH_FCM_SERVER_KEY', '')
FCM_SENDER_ID = getattr(settings, 'DJANGO_INFOPUSH_FCM_SENDER_ID', '')

# how many processes to use in a pushsend management command for parallel push
# 1 disables multiprocessing
PUSHSEND_WORKERS = int(getattr(settings, 'DJANGO_INFOPUSH_PUSHSEND_WORKERS', 3))

# default push icon
DEFAULT_ICON_URL = getattr(settings, 'DJANGO_INFOPUSH_DEFAULT_ICON_URL', "/static/push/img/icon.png")
MIN_ICON_W = int(getattr(settings, 'DJANGO_INFOPUSH_MIN_ICON_W', 192))
MIN_ICON_H = int(getattr(settings, 'DJANGO_INFOPUSH_MIN_ICON_H', 192))
# kb, max filesize of push icon
ICON_CUSTOM_MAX_FILESIZE = int(getattr(settings, 'DJANGO_INFOPUSH_ICON_CUSTOM_MAX_FILESIZE', 25))

# big push image for push in Chrome
# best aspect ration is 3:2 for desktop Chrome
# mobile Chrome will crop it vertically a little
# https://web-push-book.gauntface.com/chapter-05/02-display-a-notification/#image
MIN_BIG_IMAGE_W = int(getattr(settings, 'DJANGO_INFOPUSH_MIN_BIG_IMAGE_W', 1023))
MIN_BIG_IMAGE_H = int(getattr(settings, 'DJANGO_INFOPUSH_MIN_BIG_IMAGE_H', 682))
# kb, max filesize for big image
BIG_IMAGE_MAX_FILESIZE = int(getattr(settings, 'DJANGO_INFOPUSH_BIG_IMAGE_MAX_FILESIZE', 100))

# web-site as an "app" in manifest.json
# https://developers.google.com/web/updates/2014/11/Support-for-installable-web-apps-with-webapp-manifest-in-chrome-38-for-Android
APP_ICON_URLS = getattr(settings, 'DJANGO_INFOPUSH_APP_ICON_URLS', ["push/img/app_icon.png",])
# optional, https://developers.google.com/web/updates/2015/08/using-manifest-to-set-sitewide-theme-color
APP_THEME_COLOR = getattr(settings, 'DJANGO_INFOPUSH_APP_THEME_COLOR', None)
# optional, https://developers.google.com/web/tools/lighthouse/audits/custom-splash-screen
APP_BACKGROUND_COLOR = getattr(settings, 'DJANGO_INFOPUSH_APP_BACKGROUND_COLOR', None)

# error threshold after which we disable push subscription
ERROR_THRESHOLD = int(getattr(settings, 'DJANGO_INFOPUSH_ERROR_THRESHOLD', 30))

# do not change, it is here for easy import of this constant
GCM_URL = 'https://android.googleapis.com/gcm/send'
FCM_URL = 'https://fcm.googleapis.com/fcm/send'

# this setting allows to disable CSRF for django_infopush views only, if needed
USE_CSRF = bool(getattr(settings, 'DJANGO_INFOPUSH_USE_CSRF', True))


# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division, print_function

import string
import random
from unittest import skipUnless
from datetime import timedelta

from django.test import TestCase, override_settings
from django.urls import reverse
from django.conf import settings
from django.utils import translation, timezone
from django.utils.six import StringIO
from django.contrib.sites.models import Site
#from django.contrib.auth import get_user_model
from django.core.management import call_command

from .settings import GCM_URL, FCM_SENDER_ID
from .models import DigestSubscription, Task


def __fake_letters(length=10):
    letters = string.ascii_letters + string.digits
    letters_len = len(letters)
    s = []
    for i in range(length):
        s.append(letters[ random.randint(0, letters_len-1) ])
    return ''.join(s)


def _new_subscription_obj_for_test(is_active=True):
    s = DigestSubscription()
    s.endpoint = 'https://updates.push.services.mozilla.com/wpush/v1/' + __fake_letters(30)
    s.timezone = settings.TIME_ZONE
    s.is_active = is_active
    if is_active:
        s.activated_at = timezone.now()
    s.save()
    return s


def _new_task_obj_for_test(is_active=True, run_at=None, started_at=None, done_at=None):
    t = Task()
    t.title = __fake_letters(10)
    t.message = __fake_letters(50)
    t.url = 'https://' + Site.objects.all()[0].domain + reverse('push_info_disable')
    t.is_active = is_active
    if run_at is not None:
        t.run_at = run_at
    t.save()
    t.started_at = started_at
    t.done_at = done_at
    # fake task as done
    if is_active and started_at is not None and done_at is not None:
        for tz_layout in t.timezonelayout_set.all():
            tz_layout.started_at = started_at
            tz_layout.done_at = done_at
            tz_layout.save()
    t.save()
    return t


class PushTests(TestCase):
    def test_push_on_off_view_working(self):
        response = self.client.get(reverse('push_on_off'))
        self.assertContains(response, 'js-push-button')

    def test_push_info_disable_view_working(self):
        with translation.override('en'):
            response = self.client.get(reverse('push_info_disable'))
        # 3 major content sections
        self.assertContains(response, 'Chrome')
        self.assertContains(response, 'Firefox')
        self.assertContains(response, 'Chrome for Android')
    
    @skipUnless(FCM_SENDER_ID, "This test is needed only for FCM project setup, not for default VAPID.")
    def test_manifest_json_has_fcm_sender_id(self):
        response = self.client.get(reverse('push_manifest_json'))
        self.assertTrue( response.json()['gcm_sender_id'] == FCM_SENDER_ID )
    
    def test_view_saves_subscription_on_server(self):
        response = self.client.post(reverse('push_save'), {
            'endpoint': GCM_URL + '/hstwtdjsyTDSDGSU',
        })
        self.assertEqual(response.status_code, 200)
        self.assertTrue( int(response.json()['response']['id']) > 0 )
    
    def test_view_deactivates_subscription_on_server(self):
        obj = _new_subscription_obj_for_test(True)
        response = self.client.post(reverse('push_deactivate'), {
            'endpoint': obj.endpoint,
        })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(int(response.json()['response']['id']), obj.pk)
        obj.refresh_from_db()
        self.assertFalse(obj.is_active)
    
    @skipUnless(settings.SITE_ID, "Can't generate push-task obj without sites framework.")
    def test_last_notification_view(self):
        obj = _new_task_obj_for_test(True,
                                     timezone.now()-timedelta(days=3),
                                     timezone.now(), timezone.now())
        response = self.client.get(reverse('push_last_notification'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(obj.title, response.json()['notification']['title'])
        self.assertEqual(obj.message, response.json()['notification']['message'])
        self.assertTrue('url' in response.json()['notification'] and response.json()['notification']['url'])
        # last notification increases views counter for task
        obj.refresh_from_db()
        self.assertEqual(obj.views, 1)

    @skipUnless(settings.SITE_ID, "Can't generate push-task obj without sites framework.")
    def test_show_notification_view(self):
        obj = _new_task_obj_for_test(True,
                             timezone.now()-timedelta(days=3),
                             timezone.now(), timezone.now())
        response = self.client.get(reverse('push_show_notification', args=[obj.id,]))
        self.assertRedirects(response, obj.url+'?from=push')
        # show notification increases clicks counter
        obj.refresh_from_db()
        self.assertEqual(obj.clicks, 1)
    
    @skipUnless(settings.SITE_ID, "Can't generate push-task obj without sites framework.")
    def test_pushsend_management_command(self):
        subscr = _new_subscription_obj_for_test(True)
        task = _new_task_obj_for_test(True, timezone.now()-timedelta(days=3))
        out = StringIO()
        call_command('pushsend', stdout=out)
        subscr.refresh_from_db()
        # fake endpoint will have to give this subscription some error points
        self.assertTrue(subscr.errors > 0 or not subscr.is_active)
        task.refresh_from_db()
        self.assertTrue(task.done_at is not None)
        self.assertTrue(task.started_at is not None)
    
    # @skipUnless(settings.SITE_ID, "Can't generate push-task obj without sites framework.")
    # def test_task_admin_form(self):
    #     # enter /admin (superuser so we don't have to bother with permissions)
    #     user = get_user_model().objects.create_superuser(
    #         'username', 'admin@localhost', 'password123'
    #     )
    #     self.client.force_login(user)
    #     response = self.client.post(
    #         reverse('admin:push_task_add'), 
    #         {
    #             'title': 'push title',
    #             'message': 'push message',
    #             'url': '/',
    #             'run_at': timezone.now(),
    #         }
    #     )
    #     self.assertEqual(response.status_code, 200)
    #     # total 1 task should be in test DB now
    #     self.assertEqual(Task.objects.count(), 1)
    #     self.client.logout()
